//
//  ImagePicker.swift
//  globalfoods
//
//  Created by Jigar Moradiya on 28/09/19.
//  Copyright © 2019 Jigar. All rights reserved.
//

import UIKit
import MobileCoreServices

public protocol ImagePickerDelegate: AnyObject {
    func didSelect(image: UIImage?, fileData:Data?)
}

open class ImagePicker: NSObject {
    
    private let pickerController: UIImagePickerController
    private weak var presentationController: UIViewController?
    private weak var delegate: ImagePickerDelegate?
    private var canUploadFile = false
    private var canRemovePhoto = false
    
    public init(presentationController: UIViewController, delegate: ImagePickerDelegate, canUploadFile:Bool = false, canRemovePhoto:Bool = false) {
        self.pickerController = UIImagePickerController()
        
        super.init()
        
        self.presentationController = presentationController
        self.delegate = delegate
        self.canUploadFile = canUploadFile
        self.canRemovePhoto = canRemovePhoto
        
        self.pickerController.delegate = self
        self.pickerController.allowsEditing = true
        self.pickerController.mediaTypes = ["public.image"]
    }
    
    private func action(for type: UIImagePickerController.SourceType?, title: String) -> UIAlertAction? {
        if type == nil{
            return UIAlertAction(title: title, style: .default) { [unowned self] _ in
                let types = [kUTTypePDF]
                let importMenu = UIDocumentPickerViewController(documentTypes: types as [String], in: .import)

                importMenu.delegate = self
                importMenu.modalPresentationStyle = .formSheet

                self.presentationController?.present(importMenu, animated: true)
            }
        }else{
            guard UIImagePickerController.isSourceTypeAvailable(type!) else {
                return nil
            }
            
            return UIAlertAction(title: title, style: .default) { [unowned self] _ in
                self.pickerController.sourceType = type!
                self.presentationController?.present(self.pickerController, animated: true)
            }
        }
        
    }
    
    public func present(from sourceView: UIView) {
        
        let alertController = UIAlertController(title: nil, message: nil, preferredStyle: .actionSheet)
        
        if canRemovePhoto{
            let removeAction = UIAlertAction(title: "Remove photo", style: .default) { [unowned self] _ in
                self.delegate?.didSelect(image: nil, fileData: nil)
            }
            alertController.addAction(removeAction)
        }
        
        if let action = self.action(for: .camera, title: "Take photo") {
            alertController.addAction(action)
        }
        if canUploadFile{
            if let action = self.action(for: nil, title: "Upload File") {
                alertController.addAction(action)
            }
        }
        
        if let action = self.action(for: .photoLibrary, title: "Photo library") {
            alertController.addAction(action)
        }
        
        alertController.addAction(UIAlertAction(title: "Cancel", style: .cancel, handler: nil))
        
        if UIDevice.current.userInterfaceIdiom == .pad {
            alertController.popoverPresentationController?.sourceView = sourceView
            alertController.popoverPresentationController?.sourceRect = sourceView.bounds
            alertController.popoverPresentationController?.permittedArrowDirections = [.down, .up]
        }
        
        self.presentationController?.present(alertController, animated: true)
    }
    
    private func pickerController(_ controller: UIImagePickerController, didSelect image: UIImage?) {
        controller.dismiss(animated: true, completion: nil)
        
        self.delegate?.didSelect(image: image, fileData: nil)
    }
}

extension ImagePicker: UIImagePickerControllerDelegate {
    
    public func imagePickerControllerDidCancel(_ picker: UIImagePickerController) {
        self.pickerController(picker, didSelect: nil)
    }
    
    public func imagePickerController(_ picker: UIImagePickerController,
                                      didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey: Any]) {
        guard let image = info[.editedImage] as? UIImage else {
            return self.pickerController(picker, didSelect: nil)
        }
        self.pickerController(picker, didSelect: image)
    }
}

extension ImagePicker: UINavigationControllerDelegate {
    
}

extension ImagePicker: UIDocumentPickerDelegate {
    public func documentPicker(_ controller: UIDocumentPickerViewController, didPickDocumentsAt urls: [URL]) {
        guard let myURL = urls.first else {
            return
        }
        print("import result : \(myURL)")
        controller.dismiss(animated: true, completion: nil)
        do{
            let data = try Data.init(contentsOf: myURL)
            self.delegate?.didSelect(image: nil, fileData: data)
        }catch let error{
            print(error.localizedDescription)
        }
        
    }


    public func documentPickerWasCancelled(_ controller: UIDocumentPickerViewController) {
        print("view was cancelled")
        controller.dismiss(animated: true, completion: nil)
        
    }
}
